#include <linux/gpio.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/interrupt.h>
#include <linux/cdev.h>


#define BOUTON_PRINCIPAL 138
#define BOUTON_LEFT 139
#define BOUTON_RIGHT 140

static dev_t myLab1_dev = 0;
static struct class* myLab1_class;
static struct cdev myLab1_cdev;

bool p10_val = 1;
bool p11_val = 1;
bool p12_val = 1;
char joystick[4];

// Module can take a parameter when loaded, eg. "insmod mylab.ko debug=1"
static int debug;
module_param(debug, int, 0);

// Operation prototypes
static int     dev_open(struct inode *, struct file *);
static ssize_t dev_read(struct file *, char *, size_t, loff_t *);
static int     dev_release(struct inode *, struct file *);

// The driver's file operations
static struct file_operations fops = {
	.owner = THIS_MODULE,
	.open = dev_open,
	.read = dev_read,
	.release = dev_release,
};

static irqreturn_t mylab_button_irq_handler(int irq, void *dev_id);
static irqreturn_t mylab_left_irq_handler(int irq, void *dev_id);
static irqreturn_t mylab_right_irq_handler(int irq, void *dev_id);

/**
 * Driver initialization code.
 */
static int __init myLab1_dev_init(void) {

	// TODO
	// 1) Allocate the character device (major number)
	if(alloc_chrdev_region(&myLab1_dev,1,1, "myLab1")<0){unregister_chrdev_region(myLab1_dev,1);return -1;}
	// 2) Initialize cdev structure
	cdev_init(&myLab1_cdev,&fops);
	// 3) Add the character device
	if(cdev_add(&myLab1_cdev,myLab1_dev,1)<0){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);return -1;}
	// 4) Create the class
	myLab1_class = class_create(THIS_MODULE,"myLab1");
	if(myLab1_class == NULL){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);class_destroy(myLab1_class);return -1;}
	// 5) Create the device
	struct device *myLabDevice = device_create (myLab1_class, NULL, myLab1_dev, NULL,"myLab1");
	// 6) Request the necessary GPIOs
	if(gpio_request_one(BOUTON_PRINCIPAL, GPIOF_DIR_IN,"Button")<0){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);class_destroy(myLab1_class);device_destroy(myLab1_class,myLab1_dev);return -1;}
	if(gpio_request_one(BOUTON_LEFT, GPIOF_DIR_IN,"Left")<0){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);class_destroy(myLab1_class);device_destroy(myLab1_class,myLab1_dev);return -1;}
	if(gpio_request_one(BOUTON_RIGHT, GPIOF_DIR_IN,"Right")<0){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);class_destroy(myLab1_class);device_destroy(myLab1_class,myLab1_dev);return -1;}
	gpio_direction_input (BOUTON_PRINCIPAL);
	gpio_direction_input (BOUTON_LEFT);
	gpio_direction_input (BOUTON_RIGHT);
	// 7) Register an IRQ handler per GPIO
	int pin10irq =  gpio_to_irq(BOUTON_PRINCIPAL);
	int pin11irq =  gpio_to_irq(BOUTON_LEFT);
	int pin12irq =  gpio_to_irq(BOUTON_RIGHT);
	if(request_threaded_irq(pin10irq,mylab_button_irq_handler,NULL,IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,"Button",NULL)<0){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);class_destroy(myLab1_class);device_destroy(myLab1_class,myLab1_dev);return -1;}
	if(request_threaded_irq(pin11irq,mylab_left_irq_handler,NULL,IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,"Left",NULL)<0){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);class_destroy(myLab1_class);device_destroy(myLab1_class,myLab1_dev);return -1;}
	if(request_threaded_irq(pin12irq,mylab_right_irq_handler,NULL,IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,"Right",NULL)<0){unregister_chrdev_region(myLab1_dev,1);cdev_del(&myLab1_cdev);class_destroy(myLab1_class);device_destroy(myLab1_class,myLab1_dev);return -1;}
	p10_val=gpio_get_value(BOUTON_PRINCIPAL);
	p11_val=gpio_get_value(BOUTON_LEFT);
	p12_val=gpio_get_value(BOUTON_RIGHT);
	printk(KERN_INFO "myLab1: driver initialized\n");
	return 0;
}

/**
 * This function is called when the module is unloaded.
 * Cleanup all gpios and unregister the driver.
 */
static void __exit myLab1_dev_exit(void) {
	// TODO
	// 1) Destroy the device
	device_destroy(myLab1_class,myLab1_dev);
	// 2) Destroy the class
	class_destroy(myLab1_class);
	// 3) Delete the cdev structure
	cdev_del(&myLab1_cdev);
	// 4) Unregister the character device (deallocate)
	unregister_chrdev_region(myLab1_dev,1);
	// 5) Free the IRQs
	int pin10irq =  gpio_to_irq(BOUTON_PRINCIPAL);
	int pin11irq =  gpio_to_irq(BOUTON_LEFT);
	int pin12irq =  gpio_to_irq(BOUTON_RIGHT);
	free_irq(pin10irq,NULL);
	free_irq(pin11irq,NULL);
	free_irq(pin12irq,NULL);
	// 6) Free the GPIOs
	gpio_free(BOUTON_PRINCIPAL);
	gpio_free(BOUTON_LEFT);
	gpio_free(BOUTON_RIGHT);
	printk(KERN_INFO "myLab1: driver destroyed\n");
}

/**
 * Open operation
 */
static int dev_open(struct inode *inodep, struct file *filep) {
	// TODO
	// - Nothing...
	//printk(KERN_INFO "myLab1: device opened\n");
	return 0;
}

/**
 * Read operation
 */
static ssize_t dev_read( struct file *filep, char *buffer, size_t len, loff_t *offset) {
	// TODO
	// - Read the joystick state (ie. write it to the user space buffer)
	//printk(KERN_INFO "%d %d %d\n",p10_val,p11_val,p12_val);
	//printk(KERN_INFO "myLab1: BUTTON = %d | LEFT = %d | RIGHT = %d\n",p10_val,p11_val,p12_val);
	//printk(KERN_INFO "myLab1: device read\n");
	joystick[0]=p10_val+'0';
	joystick[1]=p11_val+'0';
	joystick[2]=p12_val+'0';
	joystick[3]=0;
	return copy_to_user(buffer, joystick, 4);;
}

/**
 * Release operation
 */
static int dev_release(struct inode *inodep, struct file *filep) {
	// TODO
	// - Nothing...
	//printk(KERN_INFO "mylab: device closed\n");
	return 0;
}


/**
 * Joystick button IRQ handler
 */
static irqreturn_t mylab_button_irq_handler(int irq, void *dev_id) {
	// TODO
	// - Update joystick state when center button is pressed
	//printk(KERN_INFO "Button Pressed\n");
	p10_val=gpio_get_value(BOUTON_PRINCIPAL);
	return (irqreturn_t) IRQ_HANDLED;  // Announce that the IRQ has been handled correctly
}

/**
 * Joystick left IRQ handler
 */
static irqreturn_t mylab_left_irq_handler(int irq, void *dev_id) {
	// TODO
	// - Update joystick state when left button is pressed
	p11_val=gpio_get_value(BOUTON_LEFT);
	return (irqreturn_t) IRQ_HANDLED;  // Announce that the IRQ has been handled correctly
}

/**
 * Joystick right IRQ handler
 */
static irqreturn_t mylab_right_irq_handler(int irq, void *dev_id) {
	// TODO
	// - Update joystick state when right button is pressed
	p12_val=gpio_get_value(BOUTON_RIGHT);
	return (irqreturn_t) IRQ_HANDLED;  // Announce that the IRQ has been handled correctly
}

/**
 * A module must use the module_init() module_exit() macros from include/linux/init.h, which
 *  identify the initialization function at insertion time and the cleanup function (as
 *  listed above)
 */
module_init(myLab1_dev_init);
module_exit(myLab1_dev_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jaime Coder <jaime.coder@hesge.ch>");
MODULE_DESCRIPTION("skeleton module");
MODULE_VERSION("0.1");
