#include "spi_screen.h"

/*
 * Print the global error message and abort execution.
 */
void pabort(const char *s)
{
    perror(s);
    abort();
}

void display_image(img_t *image,spi_display_t *disp,void *gpio_base_addr)
{
  // pixel_t tmp;
  // for (int x = 0; x < image->width; x++){
  //   for (int y = 0; y < image->height; y++){
  //     tmp = image->pix2d[y][x];
  //     disp_setpix(disp,x,y,RGB(tmp.r,tmp.g,tmp.b),gpio_base_addr);
  //   }
  // }
  int cpt = 0;
  uint8_t dat[4096];
  pixel_t tmp;
  uint8_t m, n;
  uint16_t color;
  disp_setwindow(disp, 0, 0,image->width-1, image->height-1,gpio_base_addr);
  for (int i = 0; i <  image->height; i++) {
      for (int j = 0; j < image->width; j++) {
          tmp = image->pix2d[i][j];
          color=RGB(tmp.r,tmp.g,tmp.b);
          m = color >> 8;
          n = color;
          dat[cpt]=m;
          dat[cpt+1]=n;
          cpt+=2;
          if(cpt>=4096)
          {
            spi_write_data(disp,dat,gpio_base_addr,4096);
            cpt=0;
          }
      }
  }
  if(cpt !=0)
  {
    spi_write_data(disp,dat,gpio_base_addr,cpt);
  }
}

void spi_write_cmd(spi_display_t *disp, uint8_t cmd,void *gpio_base_addr)
{
    uint8_t unused[4096];

    // one byte for command
    struct spi_ioc_transfer msg = {
            .tx_buf = (unsigned long) &cmd,
            .rx_buf = (unsigned long) unused,
            .len = 1,
            .delay_usecs = SPI_CMD_DELAY,
            .speed_hz = spi_speed,
            .bits_per_word = spi_bits,
    };

    // Set data/cmd pin to 0
    set_cmd_data(0, gpio_base_addr);

    // SPI_IOC_MESSAGE(1) indicates we are sending 1 message (the message "cmd"),
    // but we could send any number of messages at once
    // More details in kernel source: include/uapi/linux/spi/spidev.h
    if (ioctl(disp->fd_spi, SPI_IOC_MESSAGE(1), &msg) == -1)
        pabort("Failed sending spi message");
}

void spi_write_data(spi_display_t *disp, uint8_t const *data,void *gpio_base_addr, size_t data_len)
{
    uint8_t unused[4096];
    // static uint8_t to_send[4096];
    // static uint16_t cpt_byte = 0;
    // to_send[cpt_byte]=
    //printf("datalen = %d\n",data_len);
    //size_t data_len = sizeof(data);

    // one byte for command
    struct spi_ioc_transfer msg = {
        .tx_buf = (unsigned long) data,
        .rx_buf = (unsigned long) unused,
        .len = data_len,      // maximum data length is 4096 bytes
        .delay_usecs = SPI_CMD_DELAY,
        .speed_hz = spi_speed,
        .bits_per_word = spi_bits,
    };

    // Set data/cmd pin to 1
    set_cmd_data(1, gpio_base_addr);

    // SPI_IOC_MESSAGE(1) indicates we are sending 1 message (the message "cmd"),
    // but we could send any number of messages at once
    // More details in kernel source: include/uapi/linux/spi/spidev.h
    if (ioctl(disp->fd_spi, SPI_IOC_MESSAGE(1), &msg) == -1)
        pabort("Failed sending spi message");
}

void disp_setwindow(spi_display_t *disp, int x0, int y0, int x1, int y1, void *gpio_base_addr)
{
    assert(!(x1 > XRES-1 || y1 > YRES-1 || x0 > x1 || y0 > y1));

	int YSH = y0 >> 8;
	int YSL = y0;
	int YEH = y1 >> 8;
	int YEL = y1;

    // Set column (left and right zone)
    spi_write_cmd(disp,0x2A,gpio_base_addr);
    { uint8_t dat[] = { 0x00, x0, 0x00, x1 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

    // Set page address (top and bottom lines)
    spi_write_cmd(disp,0x2B,gpio_base_addr);
    { uint8_t dat[] = { YSH, YSL, YEH, YEL };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

    // Memory write
    spi_write_cmd(disp,0x2C,gpio_base_addr);
}

void disp_clear(spi_display_t *disp, uint16_t color, void *gpio_base_addr)
{
    int cpt = 0;
    uint8_t dat[4096];
    disp_setwindow(disp, 0, 0, XRES-1, YRES-1,gpio_base_addr);
    for (int i = 0; i < YRES; i++) {
        for (int j = 0; j < XRES; j++) {
            uint8_t m, n;
            m = color >> 8;
            n = color;
            dat[cpt]=m;
            dat[cpt+1]=n;
            cpt+=2;
            if(cpt>=4096)
            {
              spi_write_data(disp,dat,gpio_base_addr,4096);
              cpt=0;
            }
        }
    }
    if(cpt !=0)
    {
      spi_write_data(disp,dat,gpio_base_addr,cpt);
    }
}

void disp_setpix(spi_display_t *disp, int x, int y, uint16_t color, void *gpio_base_addr)
{
    disp_setwindow(disp, x, y, x, y,gpio_base_addr);
    uint8_t m, n;
    m = color >> 8;
    n = color;
    { uint8_t dat[] = { m, n };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }
}

void disp_init(spi_display_t *disp, void *gpio_base_addr)
{
    // SPI mode
    if (ioctl(disp->fd_spi, SPI_IOC_WR_MODE32, &spi_mode) == -1)
        pabort("Failed setting spi mode");

    // SPI bits per word
    if (ioctl(disp->fd_spi, SPI_IOC_WR_BITS_PER_WORD, &spi_bits) == -1)
        pabort("Failed setting spi bits per word");

    // SPI max speed in hz
    if (ioctl(disp->fd_spi, SPI_IOC_WR_MAX_SPEED_HZ, &spi_speed) == -1)
        pabort("Failed setting spi max speed");

    printf("SPI mode: 0x%x\n", spi_mode);
    printf("SPI bits per word: %d\n", spi_bits);
    printf("SPI max speed: %d Hz (%d KHz)\n", spi_speed, spi_speed / 1000);

    // Initialization sequence for the ILI9341 display (mylab2)
    spi_write_cmd(disp,0x01,gpio_base_addr);   // Software reset
    usleep(50000);     // Wait 50ms

    spi_write_cmd(disp,0x11,gpio_base_addr);
    usleep(120000);    // Wait 120ms

    spi_write_cmd(disp,0xCF,gpio_base_addr);
    { uint8_t dat[] = { 0x00, 0x83, 0x30 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xED,gpio_base_addr);
    { uint8_t dat[] = { 0x64, 0x03, 0x12, 0x81 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xE8,gpio_base_addr);
    { uint8_t dat[] = { 0x85, 0x01, 0x79 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xCB,gpio_base_addr);
    { uint8_t dat[] = { 0x39, 0x2C, 0x00, 0x34, 0x02 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xF7,gpio_base_addr);
    { uint8_t dat[] = { 0x20 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xEA,gpio_base_addr);
    { uint8_t dat[] = { 0x00, 0x00 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xC1,gpio_base_addr);             // Power control
    { uint8_t dat[] = { 0x11 };  // SAP[2:0];BT[3:0]
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xC5,gpio_base_addr);    // VCM control 1
    { uint8_t dat[] = { 0x34, 0x3D };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xC7,gpio_base_addr);    // VCM control 2
    { uint8_t dat[] = { 0xC0 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0x36,gpio_base_addr);    // Memory Access Control
    { uint8_t dat[] = { 0x08 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0x3A,gpio_base_addr);   	         // Pixel format
    { uint8_t dat[] = { 0x55 };  // 16bpp
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xB1,gpio_base_addr);   	               // Frame rate
    { uint8_t dat[] = { 0x00, 0x1D };  // 65Hz
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xB6,gpio_base_addr);    // Display Function Control
    { uint8_t dat[] = { 0x0A, 0xA2, 0x27, 0x00 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

	spi_write_cmd(disp,0xb7,gpio_base_addr);    // Entry mode
    { uint8_t dat[] = { 0x07 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0xF2,gpio_base_addr);    // 3Gamma Function Disable
    { uint8_t dat[] = { 0x08 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0x26,gpio_base_addr);    // Gamma curve selected
    { uint8_t dat[] = { 0x01 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

	spi_write_cmd(disp,0xE0,gpio_base_addr);    // positive gamma correction
    { uint8_t dat[] = { 0x1F, 0x1A, 0x18, 0x0A, 0x0F, 0x06, 0x45, 0x87, 0x32, 0x0A, 0x07, 0x02, 0x07, 0x05, 0x00 };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

	spi_write_cmd(disp,0xE1,gpio_base_addr);    // negamma correction
    { uint8_t dat[] = { 0x00, 0x25, 0x27, 0x05, 0x10, 0x09, 0x3A, 0x78, 0x4D, 0x05, 0x18, 0x0D, 0x38, 0x3A, 0x1F };
    spi_write_data(disp,dat,gpio_base_addr,sizeof(dat)); }

 	spi_write_cmd(disp,0x11,gpio_base_addr);    // Exit sleep
 	usleep(120000);

 	spi_write_cmd(disp,0x29,gpio_base_addr);    // Display on
 	usleep(50000);

    // Set backlight pin to 1
    set_backlight(1, gpio_base_addr);
    printf("INIT LCD FINISH\n");
}
