#ifndef SPI_SCREEN
#define SPI_SCREEN

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>

#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>

#include "nmap_gpio_screen.h"
#include "ppm.h"

#define XRES  240
#define YRES  320

#define SPIDEVNODE "/dev/spidev32765.0"

#define SPI_CMD_DELAY 30

// 16bpp (RGB 5:6:5) pixel format:
// |15-11|10-05|04-00|
// |  R  |  G  |  B  |

#define COL8TO5(x) (x >> 3)
#define COL8TO6(x) (x >> 2)

#define RGB(r,g,b) ((COL8TO5(r)<<11)|(COL8TO6(g)<<5)|(COL8TO5(b)))

#define GREEN      RGB(0,255,0)
#define BLACK      RGB(0,0,0)
#define WHITE      RGB(255,255,255)
#define GREY       RGB(100,100,100)
#define LIGHT_RED  RGB(255,120,120)

static uint32_t spi_speed = 12000000;
static uint8_t spi_bits = 8;
static uint32_t spi_mode = 0;

typedef struct {
    int fd_spi;
    struct spi_ioc_transfer cmd;
} spi_display_t;

void display_image(img_t *image,spi_display_t *disp,void *gpio_base_addr);
void disp_init(spi_display_t *disp, void *gpio_base_addr);
void disp_setpix(spi_display_t *disp, int x, int y, uint16_t color, void *gpio_base_addr);
void disp_clear(spi_display_t *disp, uint16_t color, void *gpio_base_addr);
void disp_setwindow(spi_display_t *disp, int x0, int y0, int x1, int y1, void *gpio_base_addr);
void spi_write_data(spi_display_t *disp, uint8_t const *data,void *gpio_base_addr, size_t data_len);
void spi_write_cmd(spi_display_t *disp, uint8_t cmd,void *gpio_base_addr);
void pabort(const char *s);

#endif
